/* Copyright 1998 Dean Gaudet, all rights reserved.  Contact
 * dgaudet@arctic.org for more information.
 */

#ifndef MQUEUE_H_INCLUDED
#define MQUEUE_H_INCLUDED 1

/* implements a message queue */

/* messages must have a void * as the first structure element which
 * can be manipulated as the "next" pointer by these routines
 */

#include <prlock.h>
#include <prcvar.h>

typedef struct a_message_queue a_message_queue;
struct a_message_queue {
    PRLock *lock;
    PRCondVar *cv;
    void *head;
    void **tail;
};

/* someone else can deal with making this work for non-gnuc compilers
 */

static __inline__ a_message_queue *mqueue_create(void)
{
    a_message_queue *mq;

    mq = PR_Malloc(sizeof(*mq));
    mq->lock = PR_NewLock();
    mq->cv = PR_NewCondVar(mq->lock);
    mq->head = NULL;
    mq->tail = &mq->head;
    return mq;
}

/* doesn't deal with emptying the queue... that's the caller's
 * responsibility
 */
static __inline__ void mqueue_destroy(a_message_queue *mq)
{
    PR_DestroyCondVar(mq->cv);
    PR_DestroyLock(mq->lock);
    PR_Free(mq);
}

/* add a single element to the queue */
static __inline__ void mqueue_enqueue(a_message_queue *mq, void *elt)
{
    *(void **)elt = NULL;
    PR_Lock(mq->lock);
    *mq->tail = elt;
    mq->tail = (void **)elt;
    PR_NotifyCondVar(mq->cv);
    PR_Unlock(mq->lock);
}

/* dequeue the entire queue */
static __inline__ void *mqueue_dequeue_en_masse(a_message_queue *mq, int can_block)
{
    void *res;

    PR_Lock(mq->lock);
    for (;;) {
	res = mq->head;
	if (res) break;
	if (!can_block) {
	    PR_Unlock(mq->lock);
	    return NULL;
	}
	PR_WaitCondVar(mq->cv, PR_INTERVAL_NO_TIMEOUT);
    }
    mq->head = NULL;
    mq->tail = &mq->head;
    PR_Unlock(mq->lock);
    return res;
}

#endif
