/* Copyright 1998 Dean Gaudet, all rights reserved.  Contact
 * dgaudet@arctic.org for more information.
 */

#include "flow.h"
#include "mqueue.h"
#include <ctype.h>
#include <string.h>

/* NOTE!  Remember, that even though we may restrict what we
 * parse here, there is a full parser that we can bounce to.
 * The goal is to recognize when a message appears in a common
 * format, and handle it quickly.  If it's not in a common format
 * the request is bounced to the full/slow parser.
 *
 * To say this another way, this parser does not implement the
 * full HTTP spec, and it violates the principle of "be lenient
 * in what you expect".
 */

/* in this prototype we only handle a static set of files */
#define DOCROOT "/home/dgaudet/ap/apachen/htdocs/"
static a_flow flows[] = {
    { "/", 0, DOCROOT "index.html" },
    { "/index.html", 0, DOCROOT "index.html" },
    { "/6k", 0, DOCROOT "index.html" },
};

static const char month_snames[12][4] =
{
    "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
static const char day_snames[7][4] =
{
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

static void init_flows(void)
{
    int i;
    PRFileInfo finfo;
    PRExplodedTime tms;
    PRFileDesc *fd;

    for (i = 0; i < sizeof(flows)/sizeof(flows[0]); ++i) {
	fd = PR_Open(flows[i].filename, PR_RDONLY, 0);
	if (fd == NULL) {
	    log_prerror("init_flows: PR_Open(%s, PR_RDONLY)",
		flows[i].filename);
	    PR_Abort();
	}
	if (PR_GetOpenFileInfo(fd, &finfo) != PR_SUCCESS) {
	    log_prerror("init_flows: PR_GetOpenFileInfo(%s)", flows[i].filename);
	    PR_Abort();
	}
	PR_ExplodeTime(finfo.modifyTime, PR_GMTParameters, &tms);
	/* RFC date format; as strftime '%a, %d %b %Y %T GMT' */
	flows[i].headers_out.iov_base = PR_smprintf(
	    "Last-Modified: %s, %.2d %s %d %.2d:%.2d:%.2d GMT\r\n"
	    "Content-Length: %u\r\n"
	    "Content-Type: text/html\r\n\r\n",
	    day_snames[tms.tm_wday],
	    tms.tm_mday, month_snames[tms.tm_month], tms.tm_year,
	    tms.tm_hour, tms.tm_min, tms.tm_sec,
	    finfo.size);
	flows[i].headers_out.iov_len = strlen(flows[i].headers_out.iov_base);
	/* XXX: well really we'd want to mmap or sendfile or something,
	 * but this is just a prototype
	 */
	flows[i].body.iov_base = PR_Malloc(finfo.size);
	flows[i].body.iov_len = finfo.size;
	if (PR_Read(fd, flows[i].body.iov_base, finfo.size) != finfo.size) {
	    log_prerror("init_flows: PR_Read(%s)", flows[i].filename);
	    PR_Abort();
	}
	PR_Close(fd);
	flows[i].url_len = strlen(flows[i].url);
    }
}

static int flow_request(a_request *r, a_message_queue *mq)
{
    int i;
    size_t url_len;

    /* example of a header we don't support */
    if (r->hdrs[HDR_if_modified_since].start) return -1;

    url_len = r->url.end - r->url.start;
    for (i = 0; i < sizeof(flows)/sizeof(flows[0]); ++i) {
	if (flows[i].url_len != url_len) continue;
	if (memcmp(flows[i].url, r->url.start, url_len)) continue;
	/* found: hand it to the response thread */
	r->flow = &flows[i];
	mqueue_enqueue(mq, r);
	return 0;
    }
    /* not found */
    return -1;
}

#define CR	'\015'
#define LF	'\012'

static __inline__ a_request_header_name lookup_header(const char *start, const char *end)
{
    switch (end - start) {
    case 4:
	if (!memcmp(start, "Host", 4)) return HDR_host;
	return HDR_bogus;
    case 6:
	if (!memcmp(start, "Accept", 6)) return HDR_accept;
	if (!memcmp(start, "Pragma", 6)) return HDR_pragma;
	return HDR_bogus;
    case 10:
	if (!memcmp(start, "Connection", 10)) return HDR_connection;
	if (!memcmp(start, "User-Agent", 10)) return HDR_user_agent;
	return HDR_bogus;
    case 14:
	if (!memcmp(start, "Accept-Charset", 14)) return HDR_accept_charset;
	return HDR_bogus;
    case 16:
	if (!memcmp(start, "Accept-Language", 16)) return HDR_accept_language;
	return HDR_bogus;
    case 18:
	if (!memcmp(start, "If-Modified-Since", 18)) return HDR_if_modified_since;
	return HDR_bogus;
    }
    return HDR_bogus;
}


static int parse_headers(a_request *r)
{
    char *p;
    char *pe;
    char *t;

    p = r->request_buf;
    pe = r->end_request_buf;

    /* there is room for four sentinels */
    pe[0] = CR;
    pe[1] = ':';
    pe[2] = '0';
    pe[3] = ' ';

    /* The CERN server required an extra CRLF after POST data... but
     * we're unlikely to be handling that at all here.  We may decide
     * later that we need to handle leading CRLF*, but for now we'll
     * just bounce if it happens.
     */

    /* REF-5.1:
     *     Request-Line = Method SP Request-URI SP HTTP-Version CRLF
     * 
     * We only handle GET requests.  We accept exactly one SP, not
     * multiple, and we certainly don't accept HT.  A valid
     * Request-Line has at least 5 characters: "GET /".
     *
     * We handle HTTP/1.x only, not HTTP/0.9.
     *
     * XXX: we could use a single 32-bit comparison here in almost
     * all cases
     */
    if (pe - p < 5
	|| p[0] != 'G'
	|| p[1] != 'E'
	|| p[2] != 'T'
	|| p[3] != ' ') {
	goto bounce;
    }

    /* parse the Request-URI */
    p += 4;
    r->url.start = p;
    while (*p != ' ') {		/* there is a ' ' sentinel */
	++p;
    }
    if (p >= pe) goto bounce;
    r->url.end = p;

    /* Parse the HTTP-Version, minimum length string " HTTP/1.1\r",
     * 10 chars.  Notice that p[0] == ' '.
     */
    if (pe - p < 10
	|| p[1] != 'H'
	|| p[2] != 'T'
	|| p[3] != 'T'
	|| p[4] != 'P'
	|| p[5] != '/'
	|| p[6] != '1'
	|| p[7] != '.') goto bounce;

    /* We only distinguish between HTTP/1.0 and HTTP/1.1, anything else
     * is treated as HTTP/1.1.  We handle leading zeros in the minor rev.
     */
    p += 8;
    while (*p == '0') {		/* there is a ' ' sentinel */
	++p;
    }
    t = p;
    while (isdigit(*p)) {	/* there is a '0' sentinel */
	++p;
    }
    if (p >= pe) goto bounce;
    r->http_minor_version = (p != t);

    /* We accept CRLF only.  Any other crap (like naked LF) is punted.
     */
    if (p[0] != CR || p[1] != LF) goto bounce;
    p += 2;

    /* done the Request-Line, p points at first header */

    /* XXX: before going further we could look up the URL to find out if we've
     * even got a cached flow for it. ... but this may not be a win with
     * name-vhosts because we don't know for a while what the full url is.
     * Needs work/benchmarking.
     */

    /* parse headers */

    /* Note we don't handle header continuation lines... and it causes the
     * parsing to bounce because the subsequent lines (starting with SP or HT)
     * will be parsed as a header and won't match any known header. 
     */

    memset(r->hdrs, 0, sizeof(r->hdrs));
    for (;;) {
	a_request_header_name hdr;

	/* done headers? */
	if (p[0] == CR && p[1] == LF) break;
	t = p;
	while (*p != ':') {	/* there is a ':' sentinel */
	    ++p;
	}
	if (p >= pe) goto bounce;
	hdr = lookup_header(t, p);
	if (hdr == HDR_bogus) goto bounce;
	do {
	    ++p;
	} while (*p == ' ');	/* sentinel... */
	if (p >= pe) goto bounce;
	r->hdrs[hdr].start = p;
	do {
	    ++p;
	} while (*p != CR);	/* sentinel... */
	if (p >= pe) goto bounce;
	r->hdrs[hdr].end = p;
	++p;
	if (*p != LF) goto bounce;
	++p;
    }

    return 0;

bounce:
    log_prerror("parse_headers: parse_error, should bounce request here");
    return -1;
}

static void process_request(a_request *r, a_message_queue *mq)
{
    if (parse_headers(r)
	|| flow_request(r, mq)) {
	/* we would bounce here... */
	PR_Free(r);
	return;
    }
    /* it's been handled */
}

static void read_requests(void *listen_socket, a_message_queue *mq)
{
    /* this is a list of requests which are waiting for their
     * first read() to complete
     */
    a_request requests;
    PRInt32 n_requests;
    PRPollDesc *pds;
    PRInt32 n_pds;
    PRInt32 i;
    a_request *r;
    a_request *nr;
    PRInt32 n_ready;

    n_requests = 0;
    PR_INIT_CLIST(&requests);
    n_pds = 100;	/* XXX: just an initial guess */
    pds = PR_Malloc(sizeof(*pds) * n_pds);
    if (pds == NULL) {
	PR_Abort();
    }
    for (;;) {
	/* fill in pds, we might need to grow it */
	if (n_requests + 1 > n_pds) {
	    PR_Free(pds);
	    n_pds = n_requests + 1;
	    pds = PR_Malloc(sizeof(*pds) * n_pds);
	    if (pds == NULL) {
		PR_Abort();
	    }
	}
	pds[0].fd = listen_socket;
	pds[0].in_flags = PR_POLL_READ;
	i = 0;
	r = PR_LIST_HEAD(&requests);
	while (i < n_requests) {
	    ++i;	/* pds[0] is for the listen socket */
	    pds[i].fd = r->client;
	    pds[i].in_flags = PR_POLL_READ;
	    r = PR_NEXT_LINK(r);
	}

	n_ready = PR_Poll(pds, n_requests + 1, PR_INTERVAL_NO_TIMEOUT);
	if (n_ready <= 0) {
	    log_prerror("read_requests: PR_Poll");
	    return;
	}

	/* if there's something to accept, decrement n_ready, but don't
	 * deal with it yet
	 */
	if (pds[0].out_flags & (PR_POLL_EXCEPT|PR_POLL_ERR|PR_POLL_NVAL)) {
	    log_prerror("read_requests: error on listening socket, bye bye");
	    return;
	}
	if (pds[0].out_flags & PR_POLL_READ) {
	    --n_ready;
	}
	/* deal with stuff that's ready to be read first */
	i = 0;
	r = PR_LIST_HEAD(&requests);
	while (i < n_requests) {
	    ++i;	/* pds[0] is for the listen socket */
	    nr = PR_NEXT_LINK(r);
	    if (pds[i].out_flags & (PR_POLL_EXCEPT|PR_POLL_ERR|PR_POLL_NVAL)) {
		/* XXX: do something more graceful */
		log_prerror("read_requests: pds[i].out_flags = %u", pds[i].out_flags);
discard_r:
		PR_REMOVE_LINK(r);
		PR_Close(r->client);
		PR_Free(r);
		--n_requests;
		--n_ready;
	    }
	    else if (pds[i].out_flags & PR_POLL_READ) {
		PRInt32 amt;

		amt = PR_Recv(r->client, r->request_buf, REQUEST_MAXSIZE, 0, PR_INTERVAL_NO_WAIT);
		/* XXX: uh this is bogus, we shouldn't get a would block here, but we do! */
		if (amt < 0 && PR_GetError() != PR_WOULD_BLOCK_ERROR) {
		    log_prerror("read_requests: PR_Recv error");
		    goto discard_r;
		}
		else if (amt == 0) {
		    goto discard_r;
		}
		else if (amt > 0) {
		    r->end_request_buf = r->request_buf + amt;
		    PR_REMOVE_LINK(r);
		    process_request(r, mq);
		    --n_requests;
		}
		--n_ready;
	    }
	    if (n_ready == 0) break;
	    r = nr;
	}

	if (pds[0].out_flags & PR_POLL_READ) {
	    PRInt32 amt;

	    requests.client = PR_Accept(listen_socket, &requests.sa_client, PR_INTERVAL_NO_WAIT);
	    if (requests.client == NULL) {
		log_prerror("read_requests: PR_Accept isn't ready!");
		PR_Abort();
	    }
	    nr = PR_Malloc(sizeof(*nr));
	    if (nr == NULL) {
		PR_Abort();
	    }
	    nr->client = requests.client;
	    nr->sa_client = requests.sa_client;
#if 0
	    amt = PR_Recv(nr->client, nr->request_buf, REQUEST_MAXSIZE, 0, PR_INTERVAL_NO_WAIT);
	    if (amt < 0 && PR_GetError() != PR_WOULD_BLOCK_ERROR) {
		log_prerror("read_requests: error on new socket, discarding");
		PR_Close(nr->client);
		PR_Free(nr);
	    }
	    else if (amt <= 0) {
#endif
		PR_APPEND_LINK(nr, &requests);
		++n_requests;
#if 0
	    }
	    else {
		nr->end_request_buf = nr->request_buf + amt;
		process_request(nr, mq);
	    }
#endif
	}
    }
}

void request_loop(void *listen_socket)
{
    a_message_queue *mq;
    PRThread *t_response;

    init_flows();
    mq = mqueue_create();
    t_response = PR_CreateThread(PR_USER_THREAD,
	response_loop, mq, PR_PRIORITY_NORMAL,
	PR_LOCAL_THREAD, PR_UNJOINABLE_THREAD, 0);
    if (t_response == NULL) {
	log_prerror("request_loop: unable to create response thread");
	PR_Abort();
    }
    read_requests(listen_socket, mq);
}
